__title__ = "aiolancium"
__package__ = "aiolancium"
__summary__ = "Simplistic python asyncio REST client for the Lancium Compute REST API"
__url__ = "https://github.com/giffels/aiolancium"

__version__ = "0.2.2"
__author__ = "Manuel Giffels"
__email__ = "giffels@gmail.com"
__copyright__ = "2022 %s and Contributors" % __author__
__keywords__ = ["asyncio", "lancium", "compute client"]



from abc import ABCMeta
from abc import abstractmethod


class Proxy(metaclass=ABCMeta):
    @abstractmethod
    async def __call__(self, method_name: str, *args, **kwargs):
        return NotImplemented

    @abstractmethod
    def __getattr__(self, method_name: str):
        return NotImplemented



from .auth import Authenticator
from .decorator import AuthDecorator, ResponseDecorator
from .proxy import ApiProxy
from .utilities.utilities import upload_helper

from typing import Optional


class LanciumClient(object):
    def __init__(self, api_url: str, auth: Authenticator, timeout: int = 60) -> None:
        self.api_url = api_url
        self.auth = auth
        self.api_proxy = ApiProxy(api_url=api_url, timeout=timeout)

    def __getattr__(self, item):
        return AuthDecorator(
            ResponseDecorator(getattr(self.api_proxy, item)), self.auth
        )

    async def download_file_helper(
        self, path: str, destination: str, job_id: Optional[int] = None
    ):
        if job_id:
            # job output returned as string, needs to be converted into bytes first
            content = (await self.jobs.download_job_output(job_id, path)).encode()
        else:
            content = await self.data.get_data(path)

        with open(destination, "wb") as f:
            f.write(content)

    async def upload_image_helper(
        self, path, source, name, source_type="singularity_image", chunk_size=32000000
    ):
        await upload_helper(
            awaitable_create_method=self.images.create_image,
            awaitable_upload_method=self.images.upload_image_file,
            path=path,
            source=source,
            source_type=source_type,
            chunk_size=chunk_size,
            name=name,
        )

    async def upload_file_helper(self, path, source, force=True, chunk_size=32000000):
        await upload_helper(
            awaitable_create_method=self.data.create_data_item,
            awaitable_upload_method=self.data.upload_data_file,
            path=path,
            source=source,
            source_type="file",
            chunk_size=chunk_size,
            force=force,
        )



import jsonref
import re


class OpenApiParser(object):
    def __init__(self, filename):
        self.in_path_pattern = re.compile(r"{(.*?)}")
        with open(filename, "r") as f:
            self.open_api = jsonref.load(f)

    def get_actions(self):
        resources = {}
        for path in self.paths:
            resource_group = self.get_resource_group(path)
            for method in self.get_methods(path):
                resources.setdefault(resource_group, {}).setdefault(
                    "actions", {}
                ).update(
                    {
                        self.get_operation_id(path, method): {
                            "url": self.get_substituted_path(path),
                            "method": method.upper(),
                            "content-type": self.get_content_type(path, method),
                            "parameters": self.get_parameters(path, method),
                        }
                    }
                )
        return resources

    @property
    def paths(self):
        return self.open_api["paths"].keys()

    def get_methods(self, path):
        return self.open_api["paths"][path].keys()

    def get_method_detail(self, path, method, detail):
        return self.open_api["paths"][path][method][detail]

    def get_operation_id(self, path, method):
        return self.get_method_detail(path, method, "operationId")

    def get_parameters(self, path, method):
        return_parameters = {}
        try:
            for parameter in self.get_method_detail(path, method, "parameters"):
                return_parameters.setdefault(parameter["in"], []).append(
                    parameter["name"]
                )
            return return_parameters
        except KeyError:
            return return_parameters

    def get_content_type(self, path, method):
        for key in self.get_method_detail(path, method, "requestBody")["content"]:
            return {"Content-Type": key}
        else:
            return {}

    @staticmethod
    def get_resource_group(path):
        return path.split("/")[1]

    def get_substituted_path(self, path):
        return self.in_path_pattern.sub("{}", path)



from .interfaces import Proxy
from .resources import lancium_resources
from .utilities.utilities import extract_kwargs

from simple_rest_client.api import API
from simple_rest_client.resource import BaseResource
from simple_rest_client.resource import AsyncResource

from functools import partial
from inspect import signature
from urllib.parse import quote_plus
from typing import Dict

import json


class ApiProxy(Proxy):
    def __init__(self, api_url: str, timeout: int):
        self.api = API(
            api_root_url=api_url,
            json_encode_body=False,
            headers={
                "Accept": "application/json",
            },
            timeout=timeout,
        )

        for resource in lancium_resources:
            self.api.add_resource(**resource)

    def __call__(self, method_name: str, *args, **kwargs):
        raise TypeError(f"{self.__class__.__name__} object is not callable!")

    def __getattr__(self, item):
        return ResourceProxy(resource=getattr(self.api, item))


class ResourceProxy(Proxy):
    pass_through_kwargs = [
        arg for arg in signature(BaseResource.__init__).parameters if arg != "self"
    ]

    def __init__(self, resource: AsyncResource) -> None:
        self.resource = resource

    async def __call__(self, method_name: str, *args, **kwargs):
        # Split keyword arguments into pass through arguments the `BaseResource` object
        # is expecting, the parameters the API is expecting and put anything else into
        # the body of the API call
        header_parameters = extract_kwargs(
            self.resource.actions.get(method_name)["parameters"].get("header", []),
            kwargs,
        )
        content_type = self.resource.actions.get(method_name)["content-type"]
        for header_parameter in (header_parameters, content_type):
            kwargs.setdefault("headers", {}).update(header_parameter)

        query_parameters = self.query_param_encode(
            extract_kwargs(
                self.resource.actions.get(method_name)["parameters"].get("query", []),
                kwargs,
            )
        )
        kwargs.setdefault("params", {}).update(query_parameters)

        path_parameters = extract_kwargs(
            self.resource.actions.get(method_name)["parameters"].get("path", []), kwargs
        )

        pass_through_kwargs = extract_kwargs(self.pass_through_kwargs, kwargs)

        # Check if there is a dedicated body keyword argument, otherwise consider
        # all other keyword arguments as method body and encode it properly
        body = kwargs.get("body", kwargs)

        if body and content_type["Content-Type"] == "application/json":
            # JSON encode body if content type is application/json
            body = json.dumps(body)

        awaitable_method = getattr(self.resource, method_name)
        return await awaitable_method(
            *args, *(path_parameters.values()), **pass_through_kwargs, body=body
        )

    def __getattr__(self, method_name: str):
        return partial(self, method_name)

    @staticmethod
    def query_param_encode(parameters: Dict) -> Dict:
        return {key: quote_plus(value) for key, value in parameters.items()}



import aiohttp
import jwt
from time import time


class Authenticator(object):
    def __init__(self, api_key: str) -> None:
        self.data = {"Authorization": f"Bearer {api_key}"}
        self.token = None
        self.token_expires_on = 0
        self._url = "https://portal.lancium.com/api/v1/access_tokens"

    def get_expiry_date(self) -> int:
        decoded_token = jwt.decode(
            self.token.split()[1],  # token contains "Bearer <Token>"
            algorithms=["HS256"],
            options={"verify_signature": False, "verify_aud": False},
        )
        return decoded_token["exp"]

    async def get_token(
        self,
    ) -> str:  # Should be replaced by async properties once available
        if not (self.token and self.is_token_valid):
            async with aiohttp.ClientSession(raise_for_status=True) as session:
                async with session.post(url=self._url, headers=self.data) as response:
                    self.token = response.headers.get("Authorization")
                    self.token_expires_on = self.get_expiry_date()
        return self.token

    @property
    def is_token_valid(self) -> float:
        return (
            self.token_expires_on - time() > 60
        )  # token should be at least valid for 60 s



from .openapi import OpenApiParser

from simple_rest_client.resource import AsyncResource

import os


lancium_resources = []

parser = OpenApiParser(f"{os.path.dirname(os.path.realpath(__file__))}/api.json")

for resource_group, actions in parser.get_actions().items():
    resource_class = type(
        f"Lancium{resource_group.capitalize()}Resource", (AsyncResource,), actions
    )
    lancium_resources.append(
        dict(resource_name=resource_group, resource_class=resource_class)
    )

__all__ = ["lancium_resources"]



from .auth import Authenticator
from .interfaces import ProxyDecorator, Proxy
from .utilities.utilities import extract_kwargs, get_method_name

from time import sleep
from typing import Callable


class AsynchronousResponseDecorator(ProxyDecorator):
    async def __call__(self, awaitable_method: Callable, *args, **kwargs):
        result = await awaitable_method(*args, **kwargs)
        if get_method_name(awaitable_method) in (
            "get_data",
            "download_job_output",
        ):
            while result.status_code == 202:
                sleep(1)
                result = await awaitable_method(*args, **kwargs)
        return result


class AuthDecorator(ProxyDecorator):
    def __init__(self, proxy: Proxy, auth: Authenticator):
        super().__init__(proxy)
        self.auth = auth

    async def __call__(self, awaitable_method: Callable, *args, **kwargs):
        kwargs.setdefault("headers", {}).update(
            {"Authorization": await self.auth.get_token()}
        )
        return await awaitable_method(*args, **kwargs)


class ResponseDecorator(ProxyDecorator):
    _file_info_kwargs = (
        "X-Object-Type",
        "X-Last-Modified",
        "X-Date-Created",
        "X-File-Size",
    )

    def __init__(self, proxy: Proxy):
        super().__init__(AsynchronousResponseDecorator(proxy))

    async def __call__(self, awaitable_method: Callable, *args, **kwargs):
        result = await awaitable_method(*args, **kwargs)
        # get_file_info is the only function where data is returned via the header
        if get_method_name(awaitable_method) == "get_file_info":
            return extract_kwargs(self._file_info_kwargs, result.headers)
        return result.body



from functools import partial
from hashlib import md5

from typing import Iterable

import os


async def upload_helper(
    awaitable_create_method,
    awaitable_upload_method,
    path,
    source,
    source_type,
    chunk_size,
    **kwargs,
):
    file_size = os.path.getsize(source)

    await awaitable_create_method(
        path=path, source=source, source_type=source_type, size=file_size, **kwargs
    )

    for chunk_data in read_binary_chunks_from_file(
        file_name=source, chunk_size=chunk_size
    ):
        await awaitable_upload_method(path, **chunk_data)


def extract_kwargs(keys: Iterable, kwargs):
    return {key: kwargs.pop(key) for key in keys if key in kwargs}


def get_method_name(arg):
    while isinstance(arg, partial):
        arg = arg.args[0]
    return arg


def read_binary_chunks_from_file(file_name, chunk_size):
    with open(file_name, "rb") as f:
        current_chunk_start = 0
        for chunk in iter(partial(f.read, chunk_size), b""):
            read_bytes = len(chunk)
            checksum = md5(chunk).hexdigest()
            headers = {
                "Content-Length": str(read_bytes),
                "Upload-Offset": str(current_chunk_start),
                "Upload-Checksum": checksum,
            }
            yield {"headers": headers, "body": chunk}
            current_chunk_start += read_bytes



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiolancium',
 'aiolancium.interfaces',
 'aiolancium.resources',
 'aiolancium.utilities']

package_data = \
{'': ['*']}

install_requires = \
['jsonref>=1.0.0,<2.0.0', 'pyjwt<=2.4.0', 'simple-rest-client==0.5.4']

extras_require = \
{'doc': ['sphinx-rtd-theme>=1.1.1,<2.0.0',
         'sphinxcontrib-contentui>=0.2.5,<0.3.0',
         'sphinx==4.3.1'],
 'test': ['flake8-bugbear==22.9.23',
          'black<=22.8.0',
          'aioresponses>=0.7.3,<0.8.0',
          'flake8>=5.0.4,<6.0.0']}

setup_kwargs = {
    'name': 'aiolancium',
    'version': '0.2.2',
    'description': 'AsyncIO Client for Lancium',
    'long_description': '[![Build Status](https://github.com/giffels/aiolancium/actions/workflows/unittests.yaml/badge.svg)](https://github.com/giffels/aiolancium/actions/workflows/unittests.yaml)\n[![Verification](https://github.com/giffels/aiolancium/actions/workflows/verification.yaml/badge.svg)](https://github.com/giffels/aiolancium/actions/workflows/verification.yaml)\n[![codecov](https://codecov.io/gh/giffels/aiolancium/branch/main/graph/badge.svg)](https://codecov.io/gh/giffels/aiolancium)\n[![Documentation Status](https://readthedocs.org/projects/aiolancium/badge/?version=latest)](https://aiolancium.readthedocs.io/en/latest/?badge=latest)\n[![PyPI version](https://badge.fury.io/py/aiolancium.svg)](https://badge.fury.io/py/aiolancium)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/aiolancium.svg?style=flat-square)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://github.com/giffels/aiolancium/blob/master/LICENSE)\n[![Black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n# aiolancium\n\naiolancium is a simplistic python REST client for the Lancium Compute REST API utilizing asyncio. The client itself has\nbeen developed against the [Lancium Compute REST API documentation](https://lancium.github.io/compute-api-docs/api.html).\n\n## Installation\naiolancium can be installed via [PyPi](https://pypi.org/) using\n\n```bash\npip install aiolancium\n```\n\n## How to use aiolancium\n\n```python\nfrom aiolancium.auth import Authenticator\nfrom aiolancium.client import LanciumClient\n\n# Authenticate yourself against the API and refresh your token if necessary\nauth = Authenticator(api_key="<your_api_key>")\n\n# Initialise the actual client\nclient = LanciumClient(api_url="https://portal.lancium.com/api/v1/", auth=auth)\n\n# List details about the `lancium/ubuntu` container\nawait client.images.list_image("lancium/ubuntu")\n\n# Create your hellow world first job.\njob = {"job": {"name": "GridKa Test Job",\n                   "qos": "high",\n                   "image": "lancium/ubuntu",\n                   "command_line": \'echo "Hello World"\',\n                   "max_run_time": 600}}\n\nawait client.jobs.create_job(**job)\n\n# Show all your jobs and their status in Lancium compute\njobs = await client.jobs.show_jobs()\n\nfor job in jobs["jobs"]:\n    # Retrieve the stdout/stdin output of your finished jobs\n    await client.jobs.download_job_output(job["id"], "stdout.txt")\n    await client.jobs.download_job_output(job["id"], "stderr.txt")\n    \n    # or download them to disk\n    await client.download_file_helper("stdout.txt", "stdout.txt", job["id"])\n    await client.download_file_helper("stderr.txt", "stderr.txt", job["id"])\n\n# Delete all your jobs in Lancium compute\nfor job in jobs["jobs"]:\n    await client.jobs.delete_job(id=job["id"])\n```\n\nIn order to simplify file uploads and downloads to/from the Lancium compute platform, an upload/download helper method \nhas been added to the client. \nThe upload helper takes care of reading a file in binary format and uploading it in 32 MB chunks (default) to the \nLancium persistent storage. The download helper downloads a file from the Lancium persistent storage to the local disks.\nThe download helper also supports the download of jobs outputs (stdout.txt, stderr.txt) to local disk (see example \nabove).\nUnfortunately, streaming of data is not support by the underlying `simple-rest-client`. Thus, the entire file is \ndownloaded to memory before writing to the disk.\n\n```python\nfrom aiolancium.auth import Authenticator\nfrom aiolancium.client import LanciumClient\n\n# Authenticate yourself against the API and refresh your token if necessary\nauth = Authenticator(api_key="<your_api_key>")\n\n# Initialise the actual client\nclient = LanciumClient(api_url="https://portal.lancium.com/api/v1/", auth=auth)\n\n# Upload /bin/bash to /test on the Lancium persistent storage\nawait client.upload_file_helper(path="test", source="/bin/bash")\n\n# Get information about the uploaded file\nawait client.data.get_file_info("/test")\n\n# Download the file again\nawait client.download_file_helper("/test", destination="test_downloaded_again")\n\n# Delete the uploaded file again, the \narg = {"file-path": "/test"}\nawait client.data.delete_data_item(**arg)\n\n# Alternative approach to delete the uploaded file\nawait client.data.delete_data_item("/test")\n```\n',
    'author': 'Manuel Giffels',
    'author_email': 'giffels@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



from ._proxy import Proxy

from abc import ABCMeta
from abc import abstractmethod
from functools import partial
from typing import Callable


class ProxyDecorator(Proxy, metaclass=ABCMeta):
    def __init__(self, proxy: Proxy):
        self.proxy = proxy

    @abstractmethod
    async def __call__(self, awaitable_method: Callable, *args, **kwargs):
        return NotImplemented

    def __getattr__(self, method_name: str):
        return partial(self, getattr(self.proxy, method_name))
